"""Unit tests for reviewboard.accounts.service_accounts.ServiceAccount.

Version Added:
    8.1
"""

from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING

from django.contrib.auth.models import User
from django.db.models import Q
from django.utils import timezone
from django_assert_queries import assert_queries

from reviewboard.accounts.errors import ServiceAccountUserError
from reviewboard.accounts.models import Profile
from reviewboard.accounts.service_accounts import ServiceAccount
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from django_assert_queries import ExpectedQueries


class ServiceAccountTests(TestCase):
    """Unit tests for ServiceAccount.

    Version Added:
        8.1
    """

    def test_init_with_minimum_subclass(self) -> None:
        """Testing ServiceAccount.__init__ with minimum subclass"""
        class MyServiceAccount(ServiceAccount):
            service_account_id = 'my-service-account'

        self.assertAttrsEqual(
            MyServiceAccount(),
            {
                '_claim_username': False,
                '_user': None,
                'api_token_expiration_secs': 60 * 60 * 24 * 30,
                'api_token_policy': None,
                'api_token_version': 1,
                'avatar_urls': None,
                'email': 'noreply@example.com',
                'name': 'Service Account',
                'preferred_username': 'my-service-account',
                'profile_version': 1,
                'service_account_id': 'my-service-account',
            })

    def test_init_with_minimum_init(self) -> None:
        """Testing ServiceAccount.__init__ with minimum init"""
        self.assertAttrsEqual(
            ServiceAccount(service_account_id='my-service-account'),
            {
                '_claim_username': False,
                '_user': None,
                'api_token_expiration_secs': 60 * 60 * 24 * 30,
                'api_token_policy': None,
                'api_token_version': 1,
                'avatar_urls': None,
                'email': 'noreply@example.com',
                'name': 'Service Account',
                'preferred_username': 'my-service-account',
                'profile_version': 1,
                'service_account_id': 'my-service-account',
            })

    def test_init_with_subclass_all(self) -> None:
        """Testing ServiceAccount.__init__ with overriding all attributes
        in subclass
        """
        class MyServiceAccount(ServiceAccount):
            service_account_id = 'my-super-service-account'

            name = 'My Super Service Account'
            email = 'super@example.com'
            preferred_username = 'super'
            profile_version = 2
            avatar_urls = {
                '1x': '/super.png',
                '2x': '/super@2x.png',
            }

            api_token_policy = {
                'resources': {
                    '*': {
                        'allow': ['DELETE'],
                    },
                },
            }
            api_token_expiration_secs = 60 * 60 * 24 * 120
            api_token_version = 2

        self.assertAttrsEqual(
            MyServiceAccount(),
            {
                '_claim_username': False,
                '_user': None,
                'api_token_expiration_secs': 60 * 60 * 24 * 120,
                'api_token_policy': {
                    'resources': {
                        '*': {
                            'allow': ['DELETE'],
                        },
                    },
                },
                'api_token_version': 2,
                'avatar_urls': {
                    '1x': '/super.png',
                    '2x': '/super@2x.png',
                },
                'email': 'super@example.com',
                'name': 'My Super Service Account',
                'preferred_username': 'super',
                'profile_version': 2,
                'service_account_id': 'my-super-service-account',
            })

    def test_init_with_override_all(self) -> None:
        """Testing ServiceAccount.__init__ with overriding all attributes
        during init
        """
        class MyServiceAccount(ServiceAccount):
            service_account_id = 'my-boring-service-account'

            name = 'My Boring Service Account'
            email = 'boring@example.com'
            preferred_username = 'boring'
            profile_version = 1
            avatar_urls = {
                '1x': '/boring.png',
            }

            api_token_policy = {
                'resources': {
                    '*': {
                        'block': ['*'],
                    },
                },
            }
            api_token_expiration_secs = 35
            api_token_version = 1

        self.assertAttrsEqual(
            MyServiceAccount(
                service_account_id='my-super-service-account',
                name='My Super Service Account',
                claim_username=True,
                email='super@example.com',
                preferred_username='super',
                profile_version=2,
                avatar_urls={
                    '1x': '/super.png',
                    '2x': '/super@2x.png',
                },
                api_token_policy={
                    'resources': {
                        '*': {
                            'allow': ['DELETE'],
                        },
                    },
                },
                api_token_expiration_secs=60 * 60 * 24 * 120,
                api_token_version=2,
            ),
            {
                '_claim_username': True,
                '_user': None,
                'api_token_expiration_secs': 60 * 60 * 24 * 120,
                'api_token_policy': {
                    'resources': {
                        '*': {
                            'allow': ['DELETE'],
                        },
                    },
                },
                'api_token_version': 2,
                'avatar_urls': {
                    '1x': '/super.png',
                    '2x': '/super@2x.png',
                },
                'email': 'super@example.com',
                'name': 'My Super Service Account',
                'preferred_username': 'super',
                'profile_version': 2,
                'service_account_id': 'my-super-service-account',
            })

    def test_init_with_missing_service_account_id(self) -> None:
        """Testing ServiceAccount.__init__ with missing service_account_id"""
        message = (
            'ServiceAccount.service_account_id must be set or provided '
            'during construction.'
        )

        with self.assertRaisesMessage(ValueError, message):
            ServiceAccount()

    def test_get_user_with_cached(self) -> None:
        """Testing ServiceAccount.get_user with cached result"""
        user = self.create_user()

        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        service_account._user = user

        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

    def test_get_user_create_preferred(self) -> None:
        """Testing ServiceAccount.get_user with creating preferred username"""
        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='My Service Account',
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Initial fetch for the user',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Creation of the user',
                'model': User,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Creation of the profile',
                'model': Profile,
                'type': 'INSERT',
            },
        ]

        with assert_queries(equeries):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'My',
                'last_name': 'Service Account',
                'username': 'my-service-user',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {},
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_create_preferred_with_avatars(self) -> None:
        """Testing ServiceAccount.get_user with creating preferred username
        with avatars
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='My Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Initial fetch for the user',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Creation of the user',
                'model': User,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Creation of the profile',
                'model': Profile,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
        ]

        with assert_queries(equeries):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'My',
                'last_name': 'Service Account',
                'username': 'my-service-user',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_create_after_conflict(self) -> None:
        """Testing ServiceAccount.get_user with creating preferred username
        after conflict
        """
        # Create 2 users, one with a profile, one without.
        user = self.create_user(username='my-service-user')
        user.get_profile()

        self.create_user(username='my-service-user-1')

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='My Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Fetch of "my-service-user-1"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user-1'),
            },
            {
                '__note__': 'Fetch of "my-service-user-2"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user-2'),
            },
            {
                '__note__': 'Creation of the user',
                'model': User,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=3),
            },
            {
                '__note__': 'Creation of the profile',
                'model': Profile,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=2),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'My',
                'last_name': 'Account',
                'username': 'my-service-user-2',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user-2': service_account,
        })

    def test_get_user_with_claim(self) -> None:
        """Testing ServiceAccount.get_user with claimed username"""
        self.create_user(username='my-service-user')

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='My Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
            claim_username=True,
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Creation of the profile',
                'model': Profile,
                'type': 'INSERT',
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'My',
                'last_name': 'Account',
                'username': 'my-service-user',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_with_claim_and_profile(self) -> None:
        """Testing ServiceAccount.get_user with claimed username with
        profile
        """
        user = self.create_user(username='my-service-user')
        user.get_profile()

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='My Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
            claim_username=True,
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Update of the profile',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'My',
                'last_name': 'Account',
                'username': 'my-service-user',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_with_claim_conflict(self) -> None:
        """Testing ServiceAccount.get_user with claiming username claimed
        by another service account
        """
        self.create_user(username='my-service-user')

        service_account1 = ServiceAccount(
            service_account_id='my-service-account-1',
            preferred_username='my-service-user',
            email='service-account1@example.com',
            name='First Service Account',
            claim_username=True,
        )

        service_account2 = ServiceAccount(
            service_account_id='my-service-account-2',
            preferred_username='my-service-user',
            email='service-account2@example.com',
            name='Second Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
            claim_username=True,
        )

        claim_map: dict[str, ServiceAccount] = {
            'my-service-user': service_account1,
        }

        message = (
            'Cannot claim the service account username "my-service-user". '
            'It has already been claimed by another service account. This '
            'may be a conflict between two extensions or a configuration '
            'error. Contact support if you need assistance.'
        )

        with self.assertNumQueries(0), \
             self.assertRaisesMessage(ServiceAccountUserError, message):
            service_account2.get_user(_claim_map=claim_map)

        self.assertEqual(claim_map, {
            'my-service-user': service_account1,
        })

    def test_get_user_with_max_attempts(self) -> None:
        """Testing ServiceAccount.get_user with max attempts reached
        after too many tries
        """
        user = self.create_user(username='my-service-user')
        self.assertIsNotNone(user.get_profile())

        service_account = ServiceAccount(
            service_account_id='my-service-account-1',
            preferred_username='my-service-user',
            email='service-account1@example.com',
            name='First Service Account',
        )

        claim_map: dict[str, ServiceAccount] = {}

        message = (
            r'Failed to create a unique service account user for '
            r'"my-service-user" \(error ID [0-9a-f-]+\)\. Please contact '
            r'support\.'
        )

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
        ]

        with assert_queries(equeries):
            with self.assertRaisesRegex(ServiceAccountUserError, message):
                service_account.get_user(_max_attempts=1)

        self.assertEqual(claim_map, {})

    def test_get_user_with_id_match(self) -> None:
        """Testing ServiceAccount.get_user with service account ID match"""
        user = self.create_user(
            username='my-service-user',
            email='old@example.com',
            first_name='Old First',
            last_name='Old Last',
        )
        profile = user.get_profile()
        profile.extra_data['_profile_ver'] = 1
        profile.extra_data['service_account_id'] = 'my-service-account'
        profile.save(update_fields=('extra_data',))

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='Super Duper Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()

        # This will keep old state, since the profile version didn't change.
        self.assertAttrsEqual(
            user,
            {
                'email': 'old@example.com',
                'first_name': 'Old First',
                'last_name': 'Old Last',
                'username': 'my-service-user',
            })
        self.assertTrue(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': True,
                    'settings': {},
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_with_id_match_new_profile_ver(self) -> None:
        """Testing ServiceAccount.get_user with service account ID match
        and new profile version
        """
        user = self.create_user(username='my-service-user')
        profile = user.get_profile()
        profile.extra_data['_profile_ver'] = 1
        profile.extra_data['service_account_id'] = 'my-service-account'
        profile.save(update_fields=('extra_data',))

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='Super Duper Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
            profile_version=2,
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Update of the profile',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'Super Duper',
                'last_name': 'Service Account',
                'username': 'my-service-user',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 2,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user': service_account,
        })

    def test_get_user_with_id_match_scan(self) -> None:
        """Testing ServiceAccount.get_user with service account ID match
        after user scan
        """
        self.create_user(username='my-service-user')

        user = self.create_user(
            username='my-service-user-1',
            email='old@example.com',
            first_name='Old First',
            last_name='Old Last',
        )
        user.set_password('ohboy')

        profile = user.get_profile()
        profile.extra_data['_profile_ver'] = 1
        profile.extra_data['service_account_id'] = 'my-service-account'
        profile.save(update_fields=('extra_data',))

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='Super Duper Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Fetch of "my-service-user-1"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user-1'),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()

        # This will keep old state, since the profile version didn't change.
        self.assertAttrsEqual(
            user,
            {
                'email': 'old@example.com',
                'first_name': 'Old First',
                'last_name': 'Old Last',
                'username': 'my-service-user-1',
            })
        self.assertTrue(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 1,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': True,
                    'settings': {},
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user-1': service_account,
        })

    def test_get_user_with_id_match_scan_new_profile_ver(self) -> None:
        """Testing ServiceAccount.get_user with service account ID match
        after user scan and new profile version
        """
        self.create_user(username='my-service-user')

        user = self.create_user(username='my-service-user-1')
        profile = user.get_profile()
        profile.extra_data['_profile_ver'] = 1
        profile.extra_data['service_account_id'] = 'my-service-account'
        profile.save(update_fields=('extra_data',))

        service_account = ServiceAccount(
            service_account_id='my-service-account',
            preferred_username='my-service-user',
            email='service-account@example.com',
            name='Super Duper Service Account',
            avatar_urls={
                '1x': '/user.png',
                '2x': '/user@2x.png',
            },
            profile_version=2,
        )

        claim_map: dict[str, ServiceAccount] = {}

        equeries: ExpectedQueries = [
            {
                '__note__': 'Fetch of "my-service-user"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user'),
            },
            {
                '__note__': 'Fetch of "my-service-user-1"',
                'model': User,
                'select_related': {'profile'},
                'where': Q(username='my-service-user-1'),
            },
            {
                '__note__': 'Population of user state',
                'model': User,
                'type': 'UPDATE',
                'where': Q(pk=2),
            },
            {
                '__note__': 'Update of the profile',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
            {
                '__note__': 'Population of the avatar',
                'model': Profile,
                'type': 'UPDATE',
                'where': Q(pk=1),
            },
        ]

        with assert_queries(equeries, with_tracebacks=True):
            user = service_account.get_user(_claim_map=claim_map)

        self.assertIsNotNone(user)

        user.refresh_from_db()
        self.assertAttrsEqual(
            user,
            {
                'email': 'service-account@example.com',
                'first_name': 'Super Duper',
                'last_name': 'Service Account',
                'username': 'my-service-user-1',
            })
        self.assertFalse(user.has_usable_password())

        with self.assertNumQueries(0):
            self.assertAttrsEqual(
                user.get_profile(), {
                    'extra_data': {
                        '_profile_ver': 2,
                        'service_account_id': 'my-service-account',
                    },
                    'should_send_email': False,
                    'settings': {
                        'avatars': {
                            'avatar_service_id': 'url',
                            'configuration': {
                                'url': {
                                    '1x': '/user.png',
                                    '2x': '/user@2x.png',
                                },
                            },
                        },
                    },
                })

        # Another fetch should use cache.
        with self.assertNumQueries(0):
            self.assertIs(service_account.get_user(), user)

        self.assertEqual(claim_map, {
            'my-service-user-1': service_account,
        })

    def test_get_api_token_with_create(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token"""
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()
        api_token = service_account.get_api_token(local_site=None)

        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })

    def test_get_api_token_with_create_and_policy(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token
        and policy
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
            api_token_policy={
                'resources': {
                    '*': {
                        'allow': ['DELETE'],
                    },
                }
            },
        )
        user = service_account.get_user()
        api_token = service_account.get_api_token(local_site=None)

        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {
                    'resources': {
                        '*': {
                            'allow': ['DELETE'],
                        },
                    }
                },
            })

    def test_get_api_token_with_create_and_deprecated(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token
        and existing token deprecated
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()

        # Create one token, which we'll mark deprecated.
        existing_api_token = service_account.get_api_token(local_site=None)
        existing_api_token.token_generator_id = 'legacy_sha1'
        existing_api_token.save(update_fields=('token_generator_id',))

        # Now create the new one.
        api_token = service_account.get_api_token(local_site=None)

        self.assertNotEqual(api_token, existing_api_token)
        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })

    def test_get_api_token_with_create_and_expired(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token
        and existing expired token
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()

        # Create one token, which we'll mark expired.
        existing_api_token = service_account.get_api_token(local_site=None)
        existing_api_token.expires = timezone.now() - timedelta(days=1)
        existing_api_token.save(update_fields=('expires',))

        # Now create the new one.
        api_token = service_account.get_api_token(local_site=None)

        self.assertNotEqual(api_token, existing_api_token)
        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })

    def test_get_api_token_with_create_and_expires_soon(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token
        and existing expired token
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()

        # Create one token, which we'll mark as expiring soon.
        existing_api_token = service_account.get_api_token(local_site=None)
        existing_api_token.expires = \
            timezone.now() + timedelta(seconds=60 * 60)
        existing_api_token.save(update_fields=('expires',))

        # Now create the new one.
        api_token = service_account.get_api_token(local_site=None)

        self.assertNotEqual(api_token, existing_api_token)
        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })

    def test_get_api_token_with_create_and_version_change(self) -> None:
        """Testing ServiceAccount.get_api_token with creating new token
        and version change
        """
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()

        # Create one token, which we'll soon replace.
        existing_api_token = service_account.get_api_token(local_site=None)

        # Now create the new one.
        service_account.api_token_version = 2
        api_token = service_account.get_api_token(local_site=None)

        self.assertNotEqual(api_token, existing_api_token)
        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v2.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })

    def test_get_api_token_with_existing(self) -> None:
        """Testing ServiceAccount.get_api_token with existing token"""
        service_account = ServiceAccount(
            service_account_id='my-service-account',
        )
        user = service_account.get_user()

        # Create one token, which we'll fetch again.
        existing_api_token = service_account.get_api_token(local_site=None)

        # Now fetch one.
        api_token = service_account.get_api_token(local_site=None)

        self.assertEqual(api_token, existing_api_token)
        self.assertAttrsEqual(
            api_token,
            {
                'note': (
                    'API token automatically created for '
                    'my-service-account_v1.'
                ),
                'local_site': None,
                'user': user,
                'policy': {},
            })



"""Add Repository.unique_together state for Django Evolution 0.7.

Version Added:
    2.0
"""

from django_evolution.mutations import ChangeMeta


MUTATIONS = [
    ChangeMeta('HostingServiceAccount', 'unique_together', []),
]



"""Views for managing licenses.

Version Added:
    8.0
"""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING

from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest, JsonResponse
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import ContextMixin, View
from djblets.features.decorators import feature_required
from djblets.util.serializers import DjbletsJSONEncoder
from typelets.django.json import SerializableDjangoJSONDict

from reviewboard.licensing.errors import LicenseActionError
from reviewboard.licensing.features import licensing_feature
from reviewboard.licensing.registry import license_provider_registry

if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse

    from reviewboard.licensing.provider import LicenseActionData


logger = logging.getLogger(__name__)


@method_decorator(
    (
        feature_required(licensing_feature),
        staff_member_required,
        csrf_protect,
    ),
    name='dispatch',
)
class LicensesView(ContextMixin, View):
    """Displays and processes information on known product licenses.

    Upon loading, the licenses will be checked for any updates and updated
    if needed.

    Actions on licenses can be performed by sending HTTP POST requests to
    this view with the following form data:

    ``action``:
        The name of the action.

    ``action_target``:
        A generated unique ID for the license for the purpose of this view.

    All actions are considered internal and are subject to change. There are
    no API stability guarantees.

    The following actions are supported:

    ``license-update-check``:
        Generates data for a license update check request to a separate
        licensing server. This is used for automatic license updates.

    ``process-license-update``:
        Process a license update payload from a separate licensing server.
        This is used for automatic license updates.

        It takes the following form data:

        ``check_request_data``:
            The license update check request data originally generated during
            the ``license-update-check`` action.

        ``check_response_data``:
            The payload from the licensing server.

    ``upload-license``:
        Uploads new data for a license, if supported by the License Provider.

        It takes the following form data:

        ``license_data``:
            The new license data.

    Version Added:
        8.0
    """

    def get(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP GET requests.

        This will display each known license, and handle auto-updating the
        licenses if supported by the License Providers.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple, unused):
                Unused positional arguments passed to the handler.

            **kwargs (dict, unused):
                Unused keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response for the page.
        """
        license_entries = [
            {
                'license_infos': [
                    {
                        'attrs': license_provider.get_js_license_model_data(
                           license_info=license_info,
                           request=request),
                    }
                    for license_info in license_provider.get_licenses()
                ],
                'model': license_provider.js_license_model_name,
                'view': license_provider.js_license_view_name,
            }
            for license_provider in license_provider_registry
        ]

        return render(
            request=request,
            template_name='admin/licensing.html',
            context={
                'license_entries': license_entries,
            })

    def post(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle HTTP POST requests.

        This will handle requests for license actions, dispatching out to
        the right action handler and returning a JSON payload of the
        results.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple, unused):
                Unused positional arguments passed to the handler.

            **kwargs (tuple, unused):
                Unused keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The HTTP response for the page.
        """
        action_data: LicenseActionData = {}

        # Pull out the details of the request.
        action = request.POST.get('action')

        try:
            if not action:
                raise LicenseActionError(_('Missing action data.'))

            action_target = request.POST.get('action_target')

            if not action_target:
                raise LicenseActionError(_('Missing action target.'))

            # Parse the action target.
            try:
                license_provider_id, license_id = action_target.split(':')
            except ValueError:
                raise LicenseActionError(_('Invalid action target.'))

            # See if that corresponds to a valid License Provider and license.
            license_provider = license_provider_registry.get_license_provider(
                license_provider_id)

            if license_provider is None:
                raise LicenseActionError(_('Invalid license provider.'))

            license_info = license_provider.get_license_by_id(license_id)

            if license_info is None:
                raise LicenseActionError(_(
                    'The license entry could not be found.'
                ))

            action_data_json = request.POST.get('action_data')

            if action_data_json:
                try:
                    action_data = json.loads(action_data_json)

                    if not isinstance(action_data, dict):
                        raise ValueError(_(
                            'Must be a mapping of key/value arguments.'
                        ))
                except Exception as e:
                    raise LicenseActionError(_(
                        'Invalid action data: {error}.'
                    ).format(error=e))
        except LicenseActionError as e:
            return HttpResponseBadRequest(
                json.dumps({
                    'error': str(e),
                    **e.payload,
                }),
                content_type='application/json')
        except Exception as e:
            logger.exception('Unexpected error performing license action '
                             '"%s": %s',
                             action, e)

            return HttpResponseBadRequest(
                json.dumps({
                    'error': _(
                        'Unexpected error performing license action '
                        '"{action}": {message}'
                    ).format(action=action,
                             message=e)
                }),
                content_type='application/json')

        # Invoke the action on the license provider.
        try:
            result = license_provider.call_action(
                action_id=action,
                license_info=license_info,
                action_data=action_data,
                request=request)

            return JsonResponse(result, encoder=DjbletsJSONEncoder)
        except LicenseActionError as e:
            return HttpResponseBadRequest(
                json.dumps({
                    'error': str(e),
                    **e.payload,
                }),
                content_type='application/json')
        except Exception as e:
            logger.exception('Unexpected error performing license action '
                             '"%s" for license %r on provider %r: %s',
                             action, license_info, license_provider, e)

            return HttpResponseBadRequest(
                json.dumps({
                    'error': _(
                        'Unexpected error performing license action '
                        '"{action}": {message}'
                    ).format(action=action,
                             message=e)
                }),
                content_type='application/json')



"""Search-related testing utilities."""

import tempfile
import time
from contextlib import contextmanager

import haystack
from django.conf import settings
from django.core.management import call_command
from djblets.siteconfig.models import SiteConfiguration

from reviewboard.admin.siteconfig import load_site_config


def reindex_search():
    """Rebuild the search index."""
    call_command('rebuild_index', interactive=False)

    # On Whoosh, the above is asynchronous, and we can end up trying to read
    # before we end up writing, occasionally breaking tests. We need to
    # introduce just a bit of a delay.
    #
    # Yeah, this is still sketchy, but we can't turn off the async behavior
    # or receive notification that the write has completed.
    time.sleep(0.1)


@contextmanager
def search_enabled(on_the_fly_indexing=False, backend_id='whoosh'):
    """Temporarily enable indexed search.

    Args:
        on_the_fly_indexing (bool, optional):
            Whether or not to enable on-the-fly indexing.

        backend_id (unicode, optional):
            The search backend to enable. Valid options are "whoosh" (default)
            and "elasticsearch".
    """
    siteconfig = SiteConfiguration.objects.get_current()

    old_backend_id = siteconfig.get('search_backend_id')
    old_backend_settings = siteconfig.get('search_backend_settings')

    if backend_id == 'whoosh':
        backend_settings = {
            'PATH': tempfile.mkdtemp(suffix='search-index',
                                     dir=settings.SITE_DATA_DIR),
            'STORAGE': 'file',
        }
    elif backend_id == 'elasticsearch':
        backend_settings = {
            'INDEX_NAME': 'reviewboard-tests',
            'URL': 'http://es.example.com:9200/',
        }
    else:
        raise NotImplementedError('Unexpected backend ID "%s"' % backend_id)

    siteconfig.settings.update({
        'search_enable': True,
        'search_backend_id': backend_id,
        'search_backend_settings': {
            backend_id: backend_settings,
        },
        'search_on_the_fly_indexing': on_the_fly_indexing,
    })
    siteconfig.save(update_fields=('settings',))

    load_site_config()

    try:
        yield

        haystack.connections['default'].reset_sessions()
    finally:
        siteconfig.settings.update({
            'search_enable': False,
            'search_backend_id': old_backend_id,
            'search_backend_settings': old_backend_settings,
            'search_on_the_fly_indexing': False,
        })
        siteconfig.save(update_fields=('settings',))

        load_site_config()



"""Setuptools backend for building Review Board extension packages.

Version Added:
    8.0:
    This moved from the top-level :py:mod:`reviewboard.packaging`.
"""

from __future__ import annotations

import os
import sys

from djblets.extensions.packaging.setuptools_backend import (
    BuildStaticFiles as DjbletsBuildStaticFiles,
    build_extension_cmdclass,
)
from setuptools import setup as setuptools_setup

from reviewboard.extensions.packaging.static_media import (
    RBStaticMediaBuilder,
    RBStaticMediaBuildContext,
)


class BuildStaticFiles(DjbletsBuildStaticFiles):
    """Builds static files for the extension.

    This will build the static media files used by the extension. JavaScript
    bundles will be minified and versioned. CSS bundles will be processed
    through LessCSS (if using :file:`.less` files), minified and versioned.

    Version Changed:
        8.0:
        This moved from :py:mod:`reviewboard.extensions.packaging`.
    """

    extension_entrypoint_group = 'reviewboard.extensions'
    django_settings_module = 'reviewboard.settings'
    static_media_builder_cls = RBStaticMediaBuilder
    static_media_build_context_cls = RBStaticMediaBuildContext


def setup(**setup_kwargs):
    """Build an extension package.

    Version Changed:
        8.0:
        This moved from :py:mod:`reviewboard.extensions.packaging`.

    Args:
        **setup_kwargs (dict):
            Keyword arguments to pass to the main setup method.
    """
    # Add the included conf directory so that there's a settings_local.py
    # file that can be used to package the static media.
    extensions_dir = os.path.abspath(os.path.dirname(__file__))
    sys.path.insert(0, os.path.join(extensions_dir, 'conf'))

    os.environ[str('FORCE_BUILD_MEDIA')] = str('1')

    setup_kwargs.update({
        'zip_safe': False,
        'include_package_data': True,
        'cmdclass': dict(
            build_extension_cmdclass(build_static_files_cls=BuildStaticFiles),
            **setup_kwargs.get('cmdclass', {})),
    })

    setuptools_setup(**setup_kwargs)



"""Authentication backend registry."""

from __future__ import annotations

import logging
from importlib import import_module
from typing import (Iterable, List, Optional, Sequence, TYPE_CHECKING, Type,
                    cast)

from django.conf import settings
from django.contrib.auth import get_backends
from django.utils.translation import gettext_lazy as _
from djblets.registries.registry import (ALREADY_REGISTERED, LOAD_ENTRY_POINT,
                                         NOT_REGISTERED, UNREGISTER)
from typing_extensions import TypeAlias

from reviewboard.accounts.backends.base import BaseAuthBackend
from reviewboard.accounts.backends.standard import StandardAuthBackend
from reviewboard.registries.registry import EntryPointRegistry

if TYPE_CHECKING:
    from importlib_metadata import EntryPoint


#: A type alias for a Review Board auth backend class.
#:
#: Version Added:
#:     6.0
_BaseAuthBackendClass: TypeAlias = Type[BaseAuthBackend]


logger = logging.getLogger(__name__)


_enabled_auth_backends: List[_BaseAuthBackendClass] = []
_auth_backend_setting: Optional[str] = None


class AuthBackendRegistry(EntryPointRegistry[_BaseAuthBackendClass]):
    """A registry for managing authentication backends."""

    entry_point = 'reviewboard.auth_backends'
    lookup_attrs = ('backend_id',)

    errors = {
        ALREADY_REGISTERED: _(
            '"%(item)r" is already a registered authentication backend.'
        ),
        LOAD_ENTRY_POINT: _(
            'Error loading authentication backend %(entry_point)s: %(error)s'
        ),
        NOT_REGISTERED: _(
            'No authentication backend registered with %(attr_name)s = '
            '%(attr_value)s.'
        ),
        UNREGISTER: _(
            '"%(item)r is not a registered authentication backend.'
        ),
    }

    def process_value_from_entry_point(
        self,
        entry_point: EntryPoint,
    ) -> _BaseAuthBackendClass:
        """Load the class from the entry point.

        If the class lacks a value for
        :py:attr:`~reviewboard.accounts.backends.base.BaseAuthBackend
        .backend_id`, it will be set as the entry point's name.

        Args:
           entry_point (importlib_metadata.EntryPoint):
                The entry point.

        Returns:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclass.
        """
        cls = entry_point.load()

        if not cls.backend_id:
            logger.warning('The authentication backend %r did not provide '
                           'a backend_id attribute. Setting it to the '
                           'entry point name ("%s")',
                           cls, entry_point.name)

            cls.backend_id = entry_point.name

        return cls

    def get_defaults(self) -> Iterable[_BaseAuthBackendClass]:
        """Yield the authentication backends.

        This will make sure the standard authentication backend is always
        registered and returned first.

        Yields:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclasses.
        """
        yield StandardAuthBackend

        for _module, _backend_cls_name in (
                ('ad', 'ActiveDirectoryBackend'),
                ('http_digest', 'HTTPDigestBackend'),
                ('ldap', 'LDAPBackend'),
                ('nis', 'NISBackend'),
                ('x509', 'X509Backend'),
            ):
            mod = import_module(f'reviewboard.accounts.backends.{_module}')
            yield getattr(mod, _backend_cls_name)

        yield from super().get_defaults()

    def unregister(
        self,
        backend_class: _BaseAuthBackendClass,
    ) -> None:
        """Unregister the requested authentication backend.

        Args:
            backend_class (type):
                The class of the backend to unregister. This must be a subclass
                of :py:class:`~reviewboard.accounts.backends.base
                .BaseAuthBackend`.

        Raises:
            djblets.registries.errors.ItemLookupError:
                Raised when the class cannot be found.
        """
        self.populate()

        try:
            super().unregister(backend_class)
        except self.lookup_error_class as e:
            logger.error('Failed to unregister unknown authentication '
                         'backend "%s".',
                         backend_class.backend_id)
            raise e

    def get_auth_backend(
        self,
        auth_backend_id: str,
    ) -> Optional[_BaseAuthBackendClass]:
        """Return the requested authentication backend, if it exists.

        Args:
            auth_backend_id (str):
                The unique ID of the
                :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
                class.

        Returns:
            type:
            The :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
            subclass, or ``None`` if it is not registered.
        """
        return self.get('backend_id', auth_backend_id)


def get_enabled_auth_backends() -> Sequence[_BaseAuthBackendClass]:
    """Return all authentication backends being used by Review Board.

    The returned list contains every authentication backend that Review Board
    will try, in order.

    Returns:
        list of type:
        The list of registered
        :py:class:`~reviewboard.accounts.backends.base.BaseAuthBackend`
        subclasses.
    """
    global _enabled_auth_backends
    global _auth_backend_setting

    if (not _enabled_auth_backends or
        _auth_backend_setting != settings.AUTHENTICATION_BACKENDS):
        _auth_backend_setting = settings.AUTHENTICATION_BACKENDS

        # We're casting because we control all the auth backends, and know
        # they're all actually ours.
        _enabled_auth_backends = cast(List[_BaseAuthBackendClass],
                                      list(get_backends()))

    return _enabled_auth_backends


#: Registry instance for working with available authentication backends.
auth_backends = AuthBackendRegistry()



"""Validators for diffviewer components."""

import re

from django.core.validators import RegexValidator
from django.utils.translation import gettext_lazy as _


#: The maximum length of a commit ID.
COMMIT_ID_LENGTH = 64

#: A regular expression for matching commit IDs.
COMMIT_ID_RE = re.compile(r'[A-Za-z0-9]{1,%s}' % COMMIT_ID_LENGTH)

#: A validator for commit IDs.
validate_commit_id = RegexValidator(COMMIT_ID_RE,
                                    _('Commits must be alphanumeric.'))



"""Tests for the DiffFileAttachmentResource."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from djblets.webapi.errors import DUPLICATE_ITEM, PERMISSION_DENIED
from djblets.webapi.testing.decorators import webapi_test_template

from reviewboard.attachments.models import FileAttachment
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.webapi.errors import DIFF_TOO_BIG
from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (
    diff_file_attachment_item_mimetype,
    diff_file_attachment_list_mimetype)
from reviewboard.webapi.tests.mixins import (BasicPostTestSetupState,
                                             BasicTestsMetaclass)
from reviewboard.webapi.tests.urls import (get_diff_file_attachment_item_url,
                                           get_diff_file_attachment_list_url)

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from typelets.json import JSONDict

    from reviewboard.diffviewer.models import FileDiff
    from reviewboard.reviews.models import ReviewRequest


class ResourceListTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DiffFileAttachmentResource list APIs."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'repositories/<id>/diff-file-attachments/'
    resource = resources.diff_file_attachment

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['filename'], attachment.filename)
        self.assertEqual(item_rsp['caption'], attachment.caption)
        self.assertEqual(item_rsp['mimetype'], attachment.mimetype)

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository()

        return get_diff_file_attachment_list_url(repository)

    def setup_http_not_allowed_list_test(self, user):
        repository = self.create_repository()

        return get_diff_file_attachment_list_url(repository)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        repository = self.create_repository(with_local_site=with_local_site)
        review_request = self.create_review_request(repository=repository)

        if populate_items:
            diffset = self.create_diffset(review_request=review_request,
                                          repository=repository)
            filediff = self.create_filediff(diffset)
            items = [self.create_diff_file_attachment(filediff)]
        else:
            items = []

        return (get_diff_file_attachment_list_url(repository, local_site_name),
                diff_file_attachment_list_mimetype,
                items)

    def test_get_with_mimetype(self):
        """Testing the GET repositories/<id>/diff-file-attachments/ API
        with ?mimetype=
        """
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(
            filediff,
            caption='Image',
            orig_filename='image.png',
            mimetype='image/png')
        self.create_diff_file_attachment(
            filediff,
            caption='Text',
            orig_filename='text.txt',
            mimetype='text/plain')

        rsp = self.api_get(
            get_diff_file_attachment_list_url(repository) +
            '?mimetype=image/png',
            expected_mimetype=diff_file_attachment_list_mimetype)
        assert rsp is not None
        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_file_attachments', rsp)

        attachments_rsp = rsp['diff_file_attachments']
        self.assertEqual(len(attachments_rsp), 1)
        attachment_rsp = attachments_rsp[0]
        self.assertEqual(attachment_rsp['id'], attachment.pk)
        self.assertEqual(attachment_rsp['filename'], 'image.png')
        self.assertEqual(attachment_rsp['caption'], 'Image')
        self.assertEqual(attachment_rsp['mimetype'], 'image/png')

    def test_get_with_repository_file_path(self):
        """Testing the GET repositories/<id>/diff-file-attachments/ API
        with ?repository-file-path=
        """
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff1 = self.create_filediff(diffset,
                                         source_file='/test-file-1.png',
                                         dest_file='/test-file-1.png')
        filediff2 = self.create_filediff(diffset,
                                         source_file='/test-file-2.png',
                                         dest_file='/test-file-2.png')
        attachment = self.create_diff_file_attachment(
            filediff1,
            caption='File 1',
            orig_filename='/test-file-1.png')
        self.create_diff_file_attachment(
            filediff2,
            caption='File 2',
            orig_filename='/test-file-2.png')

        rsp = self.api_get(
            get_diff_file_attachment_list_url(repository) +
            '?repository-file-path=/test-file-1.png',
            expected_mimetype=diff_file_attachment_list_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_file_attachments', rsp)

        attachments_rsp = rsp['diff_file_attachments']
        self.assertEqual(len(attachments_rsp), 1)
        attachment_rsp = attachments_rsp[0]
        self.assertEqual(attachment_rsp['id'], attachment.pk)
        self.assertEqual(attachment_rsp['filename'], '/test-file-1.png')
        self.assertEqual(attachment_rsp['caption'], 'File 1')
        self.assertEqual(attachment_rsp['mimetype'], 'image/png')

    def test_get_with_repository_revision(self):
        """Testing the GET repositories/<id>/diff-file-attachments/ API
        with ?repository-revision=
        """
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff1 = self.create_filediff(diffset,
                                         source_file='/test-file-1.png',
                                         dest_file='/test-file-1.png',
                                         source_revision='4',
                                         dest_detail='5')
        filediff2 = self.create_filediff(diffset,
                                         source_file='/test-file-2.png',
                                         dest_file='/test-file-2.png',
                                         source_revision='9',
                                         dest_detail='10')
        attachment = self.create_diff_file_attachment(
            filediff1,
            from_modified=False,
            caption='File 1',
            orig_filename='/test-file-1.png')
        self.create_diff_file_attachment(
            filediff2,
            from_modified=False,
            caption='File 2',
            orig_filename='/test-file-2.png')

        rsp = self.api_get(
            get_diff_file_attachment_list_url(repository) +
            '?repository-revision=4',
            expected_mimetype=diff_file_attachment_list_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_file_attachments', rsp)

        attachments_rsp = rsp['diff_file_attachments']
        self.assertEqual(len(attachments_rsp), 1)
        attachment_rsp = attachments_rsp[0]
        self.assertEqual(attachment_rsp['id'], attachment.pk)
        self.assertEqual(attachment_rsp['filename'], '/test-file-1.png')
        self.assertEqual(attachment_rsp['caption'], 'File 1')
        self.assertEqual(attachment_rsp['mimetype'], 'image/png')

    #
    # HTTP POST tests
    #

    def check_post_result(
        self,
        user: User,
        rsp: JSONDict,
        filediff: FileDiff,
        review_request: ReviewRequest,
        source_file: bool = None,
        *args,
        **kwargs,
    ) -> None:
        """Check the results of an HTTP POST.

        Args:
            user (django.contrib.auth.models.User):
                The user performing the request.

            rsp (dict):
                The POST response payload.

            *args (tuple):
                Positional arguments provided by
                :py:meth:`populate_post_test_objects`.

            **kwargs (dict):
                Keyword arguments provided by
                :py:meth:`populate_post_test_objects`.

        Raises:
            AssertionError:
                One of the checks failed.
        """
        self.assertIn('diff_file_attachment', rsp)
        item_rsp = rsp['diff_file_attachment']

        attachment = FileAttachment.objects.get(pk=item_rsp['id'])

        if source_file:
            self.assertEqual(attachment.repo_path, filediff.source_file)
            self.assertEqual(attachment.repo_revision,
                             filediff.source_revision)
        else:
            self.assertEqual(attachment.added_in_filediff, filediff)

        review_request.refresh_from_db()
        self.assertEqual(review_request.file_attachments_count, 1)

    def populate_post_test_objects(
        self,
        *,
        setup_state: BasicPostTestSetupState,
        create_valid_request_data: bool,
        new_file: bool = False,
        **kwargs,
    ) -> None:
        """Populate objects for a POST test.

        Args:
            setup_state (reviewboard.webapi.tests.mixins.
                         BasicPostTestSetupState):
                The setup state for the test.

            create_valid_request_data (bool):
                Whether ``request_data`` in ``setup_state`` should provide
                valid data for a POST test, given the populated objects.

            new_file (bool, optional):
                Whether to create a newly-added file.

            **kwargs (dict):
                Additional keyword arguments for future expansion.
        """
        repository = self.create_repository(
            local_site=setup_state['local_site'])
        review_request = self.create_review_request(
            local_site=setup_state['local_site'],
            submitter=setup_state['owner'],
            repository=repository,
            create_with_history=True)
        diffset = self.create_diffset(review_request)
        commit = self.create_diffcommit(diffset=diffset)

        self.assertEqual(review_request.file_attachments_count, 0)

        if new_file:
            source_revision = PRE_CREATION
        else:
            source_revision = '123'

        filediff = self.create_filediff(
            diffset=diffset,
            commit=commit,
            source_revision=source_revision,
            diff=b'',
            binary=True)

        if create_valid_request_data:
            setup_state['request_data'] = {
                'path': open(self.get_sample_image_filename(), 'rb'),
                'filediff': filediff.pk,
            }
        else:
            setup_state['request_data'] = {}

        setup_state['url'] = get_diff_file_attachment_list_url(
            repository, setup_state['local_site_name'])
        setup_state['mimetype'] = diff_file_attachment_item_mimetype
        setup_state['check_result_args'] = (filediff, review_request)

    @webapi_test_template
    def test_post_with_new_file(self) -> None:
        """Testing the POST <URL> API with a newly-added binary file"""
        resource = self.resource

        self.assertTrue(getattr(resource.create, 'login_required', False))
        self.assertTrue(getattr(resource.create, 'checks_local_site', False))

        self.load_fixtures(self.basic_post_fixtures)
        self._login_user(admin=self.basic_post_use_admin)

        setup_state = cast(
            BasicPostTestSetupState,
            self._build_common_setup_state(fixtures=self.basic_post_fixtures))
        self.populate_post_test_objects(
            setup_state=setup_state,
            create_valid_request_data=True,
            new_file=True)

        request_data = setup_state.get('request_data')

        rsp = self.api_post(setup_state['url'],
                            request_data,
                            expected_mimetype=setup_state['mimetype'],
                            expected_status=self.basic_post_success_status)

        assert rsp
        self.assertEqual(rsp['stat'], 'ok')
        self.check_post_result(setup_state['owner'],
                               rsp,
                               *setup_state.get('check_result_args', ()),
                               **setup_state.get('check_result_kwargs', {}))

    @webapi_test_template
    def test_post_source_file(self) -> None:
        """Testing the POST <URL> API with source_file=1"""
        resource = self.resource

        self.assertTrue(getattr(resource.create, 'login_required', False))
        self.assertTrue(getattr(resource.create, 'checks_local_site', False))

        self.load_fixtures(self.basic_post_fixtures)
        self._login_user(admin=self.basic_post_use_admin)

        setup_state = cast(
            BasicPostTestSetupState,
            self._build_common_setup_state(fixtures=self.basic_post_fixtures))
        self.populate_post_test_objects(
            setup_state=setup_state,
            create_valid_request_data=True)

        request_data = setup_state.get('request_data')
        request_data['source_file'] = '1'

        rsp = self.api_post(setup_state['url'],
                            request_data,
                            expected_mimetype=setup_state['mimetype'],
                            expected_status=self.basic_post_success_status)

        assert rsp
        self.assertEqual(rsp['stat'], 'ok')
        self.check_post_result(setup_state['owner'],
                               rsp,
                               source_file=True,
                               *setup_state.get('check_result_args', ()),
                               **setup_state.get('check_result_kwargs', {}))

    @webapi_test_template
    def test_post_duplicate_file(self) -> None:
        """Testing the POST <URL> API with a duplicate file"""
        resource = self.resource

        self.assertTrue(getattr(resource.create, 'login_required', False))
        self.assertTrue(getattr(resource.create, 'checks_local_site', False))

        self.load_fixtures(self.basic_post_fixtures)
        self._login_user(admin=self.basic_post_use_admin)

        setup_state = cast(
            BasicPostTestSetupState,
            self._build_common_setup_state(fixtures=self.basic_post_fixtures))
        self.populate_post_test_objects(
            setup_state=setup_state,
            create_valid_request_data=True)

        filediff = setup_state['check_result_args'][0]
        self.create_diff_file_attachment(filediff)

        request_data = setup_state.get('request_data')

        rsp = self.api_post(setup_state['url'],
                            request_data,
                            expected_status=409)

        assert rsp
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DUPLICATE_ITEM.code)

    @webapi_test_template
    def test_post_file_too_big(self) -> None:
        """Testing the POST <URL> API with a file that exceeds configured size
        limits
        """
        resource = self.resource

        self.assertTrue(getattr(resource.create, 'login_required', False))
        self.assertTrue(getattr(resource.create, 'checks_local_site', False))

        self.load_fixtures(self.basic_post_fixtures)
        self._login_user(admin=self.basic_post_use_admin)

        setup_state = cast(
            BasicPostTestSetupState,
            self._build_common_setup_state(fixtures=self.basic_post_fixtures))
        self.populate_post_test_objects(
            setup_state=setup_state,
            create_valid_request_data=True,
            new_file=True)

        request_data = setup_state.get('request_data')

        with self.siteconfig_settings({'diffviewer_max_binary_size': 2},
                                      reload_settings=False):
            rsp = self.api_post(setup_state['url'],
                                request_data,
                                expected_status=400)

        assert rsp is not None
        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], DIFF_TOO_BIG.code)
        self.assertIn('reason', rsp)
        self.assertIn('max_size', rsp)
        self.assertEqual(rsp['max_size'], 2)


class ResourceItemTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Testing the DiffFileAttachmentResource item APIs."""

    fixtures = ['test_users', 'test_scmtools']
    sample_api_url = 'repositories/<id>/diff-file-attachments/<id>/'
    resource = resources.diff_file_attachment

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['filename'], attachment.filename)
        self.assertEqual(item_rsp['caption'], attachment.caption)
        self.assertEqual(item_rsp['mimetype'], attachment.mimetype)

    def setup_http_not_allowed_item_test(self, user):
        repository = self.create_repository()
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        return get_diff_file_attachment_item_url(repository, attachment)

    def setup_http_not_allowed_list_test(self, user):
        assert self.user is not None

        repository = self.create_repository(public=False)
        repository.users.add(self.user)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        return get_diff_file_attachment_item_url(attachment, repository)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        repository = self.create_repository(with_local_site=with_local_site)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        return (get_diff_file_attachment_item_url(attachment, repository,
                                                  local_site_name),
                diff_file_attachment_item_mimetype,
                attachment)

    def test_get_with_invite_only_repo(self):
        """Testing the GET repositories/<id>/diff-file-attachments/<id>/ API
        with access to an invite-only repository
        """
        assert self.user is not None

        repository = self.create_repository(public=False)
        repository.users.add(self.user)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        rsp = self.api_get(
            get_diff_file_attachment_item_url(attachment, repository),
            expected_mimetype=diff_file_attachment_item_mimetype)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'ok')
        self.assertIn('diff_file_attachment', rsp)

        attachment_rsp = rsp['diff_file_attachment']
        self.assertEqual(attachment_rsp['id'], attachment.pk)
        self.assertEqual(attachment_rsp['filename'], attachment.filename)
        self.assertEqual(attachment_rsp['caption'], attachment.caption)
        self.assertEqual(attachment_rsp['mimetype'], attachment.mimetype)

    def test_get_with_invite_only_repo_no_access(self):
        """Testing the GET repositories/<id>/diff-file-attachments/<id>/ API
        without access to an invite-only repository
        """
        repository = self.create_repository(public=False)
        review_request = self.create_review_request(repository=repository)
        diffset = self.create_diffset(review_request=review_request,
                                      repository=repository)
        filediff = self.create_filediff(diffset)
        attachment = self.create_diff_file_attachment(filediff)

        rsp = self.api_get(
            get_diff_file_attachment_item_url(attachment, repository),
            expected_status=403)
        assert rsp is not None

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)



"""Unit tests for reviewboard.accounts.middleware.UpdateLastLoginMiddleware."""

from datetime import datetime, timedelta, timezone

from django.contrib.auth.models import User
from django.http import HttpResponse
from django.test.client import RequestFactory
from django.utils import timezone as django_timezone
from kgb import SpyAgency

from reviewboard.accounts.middleware import update_last_login_middleware
from reviewboard.testing import TestCase


class UpdateLastLoginMiddlewareTests(SpyAgency, TestCase):
    """Unit tests for UpdateLastLoginMiddleware."""

    fixtures = ['test_users']

    def setUp(self):
        super(UpdateLastLoginMiddlewareTests, self).setUp()

        self.middleware = update_last_login_middleware(
            lambda request: HttpResponse(''))
        self.user = User.objects.create(username='test-user')

        self.request = RequestFactory().get('/')
        self.request.user = self.user

        self.spy_on(django_timezone.now,
                    call_fake=lambda: datetime(year=2018, month=3, day=3,
                                               hour=19, minute=30, second=0,
                                               tzinfo=timezone.utc))

        self.now = django_timezone.now()

    def test_process_request_with_gt_30_mins(self):
        """Testing update_last_login_middleware with last login time > 30
        minutes old
        """
        self.user.last_login = self.now - timedelta(seconds=31 * 60)
        self.middleware(self.request)

        self.assertEqual(self.user.last_login, self.now)

        # Make sure this has saved.
        user = User.objects.get(pk=self.user.pk)
        self.assertEqual(user.last_login, self.now)

    def test_process_request_with_lt_30_mins(self):
        """Testing update_last_login_middleware with last login time < 30
        minutes old
        """
        cur_last_login = self.now - timedelta(seconds=15 * 60)
        self.user.last_login = cur_last_login
        self.middleware(self.request)

        self.assertEqual(self.user.last_login, cur_last_login)



"""Unit tests for reviewboard.diffviewer.views.DownloadPatchErrorBundleView."""

import kgb
from django.http import HttpResponse

from reviewboard.diffviewer.errors import PatchError
from reviewboard.diffviewer.renderers import DiffRenderer
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase


class DownloadPatchErrorBundleViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for DownloadPatchErrorBundleView."""

    fixtures = ['test_users', 'test_scmtools']

    def test_sends_404_when_no_patch_error(self):
        """Testing DownloadPatchErrorBundleView returns 404 when
        no patch error is raised by the renderer.
        """
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request=review_request)
        self.create_diffcommit(diffset=diffset)
        filediff_id = diffset.files.all()[0].pk

        # This does not raise a PatchError, so there is no patch error bundle.
        self.spy_on(DiffRenderer.render_to_response,
                    op=kgb.SpyOpReturn(HttpResponse()))

        response = self.client.get(
            local_site_reverse(
                'patch-error-bundle',
                kwargs={
                    'review_request_id': review_request.pk,
                    'revision': diffset.revision,
                    'filediff_id': filediff_id,
                }))

        self.assertEqual(response.status_code, 404)

    def test_sends_bundle_when_patch_error(self):
        """Testing DownloadPatchErrorBundleView sends a patch error bundle
        when a PatchError is raised by the renderer.
        """
        review_request = self.create_review_request(publish=True,
                                                    create_repository=True)
        diffset = self.create_diffset(review_request=review_request)
        self.create_diffcommit(diffset=diffset)
        filediff_id = diffset.files.all()[0].pk

        patch_error = PatchError(filename='filename',
                                 error_output='error_output',
                                 orig_file=b'orig_file',
                                 new_file=b'new_file',
                                 diff=b'diff',
                                 rejects=b'rejects')
        self.spy_on(DiffRenderer.render_to_response,
                    op=kgb.SpyOpRaise(patch_error))

        response = self.client.get(
            local_site_reverse(
                'patch-error-bundle',
                kwargs={
                    'review_request_id': review_request.pk,
                    'revision': diffset.revision,
                    'filediff_id': filediff_id,
                }))

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response['Content-Type'], 'application/zip')



"""An interface to a bug tracker.

Version Changed:
    8.0:
    Renamed this module from ``reviewboard.hostingsvcs.bugtracker``.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

from djblets.cache.backend import cache_memoize

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import ClassVar, Literal

    from typing_extensions import NotRequired

    from reviewboard.scmtools.models import Repository


class BugInfo(TypedDict):
    """Information about a bug.

    Version Added:
        8.0
    """

    #: The description of the bug.
    description: str

    #: The text format of the description.
    description_text_format: NotRequired[Literal['html', 'markdown', 'plain']]

    #: A one-line summary of the bug.
    summary: str

    #: The bug's status.
    status: str


class BaseBugTracker:
    """An interface to a bug tracker.

    Bug tracker subclasses are used to enable interaction with different
    bug trackers.

    Version Changed:
        8.0:
        Moved and renamed from
        ``reviewboard.hostingsvcs.bugtracker.BugTracker``.
    """

    #: The name of the bug tracker
    name: ClassVar[str | None] = None

    def get_bug_info(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        This should return a :py:class:`BugInfo` dictionary.

        This is cached for 60 seconds to reduce the number of queries to the
        bug trackers and make things seem fast after the first infobox load,
        but is still a short enough time to give relatively fresh data.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            BugInfo:
            Information about the bug.
        """
        return cache_memoize(self.make_bug_cache_key(repository, bug_id),
                             lambda: self.get_bug_info_uncached(repository,
                                                                bug_id),
                             expiration=60)

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        This should be implemented by subclasses.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            BugInfo:
            Information about the bug.
        """
        return {
            'summary': '',
            'description': '',
            'status': '',
        }

    def make_bug_cache_key(
        self,
        repository: Repository,
        bug_id: str,
    ) -> str | Sequence[str]:
        """Return a key to use when caching fetched bug information.

        Version Changed:
            8.0:
            Changed to return a list of strings for the cache key, which
            :py:func:`djblets.cache.backend.cache_memoize` will use to create a
            safely escaped cache key.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug.

        Returns:
            list of str:
            A key to use for the cache.
        """
        return [
            'repository',
            str(repository.pk),
            'bug',
            bug_id,
        ]



"""Add Group.invite_only.

Version Added:
    1.6
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Group', 'invite_only', models.BooleanField, initial=False),
]



"""Unit tests for reviewboard.extensions.hooks.NavigationBarHook."""

from django.contrib.auth.models import User
from django.template import Context, Template

from reviewboard.extensions.hooks import NavigationBarHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase


class NavigationBarHookTests(BaseExtensionHookTestCase):
    """Tests the navigation bar hooks."""

    def test_get_entries(self):
        """Testing NavigationBarHook.get_entries"""
        entry = {
            'label': 'Test Nav Entry',
            'url': 'foo-url',
        }

        hook = NavigationBarHook(extension=self.extension, entries=[entry])

        request = self.client.request()
        request.user = User(username='text')

        context = Context({
            'request': request,
            'local_site_name': 'test-site',
        })
        entries = hook.get_entries(context)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0], entry)

        t = Template(
            '{% load rb_extensions %}'
            '{% navigation_bar_hooks %}')

        self.assertEqual(
            t.render(context).strip(),
            ('<li class="rb-c-topbar__nav-item">'
             '<a href="%(url)s">%(label)s</a></li>'
             % entry))

    def test_get_entries_with_url_name(self):
        """Testing NavigationbarHook.get_entries with URL names"""
        entry = {
            'label': 'Test Nav Entry',
            'url_name': 'dashboard',
        }

        hook = NavigationBarHook(extension=self.extension, entries=[entry])

        request = self.client.request()
        request.user = User(username='text')

        context = Context({
            'request': request,
            'local_site_name': 'test-site',
        })
        entries = hook.get_entries(context)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0], entry)

        t = Template(
            '{% load rb_extensions %}'
            '{% navigation_bar_hooks %}')

        self.assertEqual(
            t.render(context).strip(),
            '<li class="rb-c-topbar__nav-item">'
            '<a href="%(url)s">%(label)s</a></li>'
            % {
                'label': entry['label'],
                'url': '/dashboard/',
            })

    def test_is_enabled_for_true(self):
        """Testing NavigationBarHook.is_enabled_for and returns true"""
        def is_enabled_for(**kwargs):
            self.assertEqual(kwargs['user'], request.user)
            self.assertEqual(kwargs['request'], request)
            self.assertEqual(kwargs['local_site_name'], 'test-site')

            return True

        entry = {
            'label': 'Test Nav Entry',
            'url': 'foo-url',
        }

        hook = NavigationBarHook(extension=self.extension, entries=[entry],
                                 is_enabled_for=is_enabled_for)

        request = self.client.request()
        request.user = User(username='text')

        context = Context({
            'request': request,
            'local_site_name': 'test-site',
        })
        entries = hook.get_entries(context)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0], entry)

        t = Template(
            '{% load rb_extensions %}'
            '{% navigation_bar_hooks %}')

        self.assertEqual(
            t.render(context).strip(),
            ('<li class="rb-c-topbar__nav-item">'
             '<a href="%(url)s">%(label)s</a></li>'
             % entry))

    def test_is_enabled_for_false(self):
        """Testing NavigationBarHook.is_enabled_for and returns false"""
        def is_enabled_for(**kwargs):
            self.assertEqual(kwargs['user'], request.user)
            self.assertEqual(kwargs['request'], request)
            self.assertEqual(kwargs['local_site_name'], 'test-site')

            return False

        entry = {
            'label': 'Test Nav Entry',
            'url': 'foo-url',
        }

        hook = NavigationBarHook(extension=self.extension, entries=[entry],
                                 is_enabled_for=is_enabled_for)

        request = self.client.request()
        request.user = User(username='text')

        context = Context({
            'request': request,
            'local_site_name': 'test-site',
        })
        entries = hook.get_entries(context)
        self.assertEqual(len(entries), 0)

        t = Template(
            '{% load rb_extensions %}'
            '{% navigation_bar_hooks %}')

        self.assertEqual(t.render(context).strip(), '')



"""Unit tests for reviewboard.reviews.models.BaseComment.

Version Added:
    8.0
"""

from __future__ import annotations

import kgb
from django.db import models

from reviewboard.reviews.models import BaseComment
from reviewboard.reviews.signals import comment_issue_status_updated
from reviewboard.testing import TestCase


class BaseCommentTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.models.BaseComment.

    Version Added:
        8.0
    """

    fixtures = ['test_users']

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        review_request = self.create_review_request(publish=True)
        review = self.create_review(review_request=review_request)
        self.review_request = review_request
        self.review_draft = review

    def test_save_kwargs(self) -> None:
        """Testing BaseComment.save passes its arguments to the parent save
        method
        """
        comment = self.create_general_comment(review=self.review_draft)

        self.spy_on(models.Model.save, owner=models.Model)

        comment.issue_opened = True
        comment.issue_status = BaseComment.OPEN

        comment.save(
            None,  # Passing `force_insert` as a positional argument.
            update_fields=['issue_opened', 'issue_status', 'timestamp'])

        self.assertSpyCalledWith(
            models.Model.save,
            force_insert=None,
            update_fields={'issue_opened', 'issue_status', 'timestamp'})

    def test_issue_updated_signal_with_created(self) -> None:
        """Testing comment creation does not emit the
        comment_issue_status_updated signal
        """
        def on_issue_status_updated(**kwargs) -> None:
            pass

        self.addCleanup(comment_issue_status_updated.disconnect,
                        on_issue_status_updated)

        comment_issue_status_updated.connect(on_issue_status_updated)
        self.spy_on(on_issue_status_updated)

        # Create a comment that opens an issue, and one that doesn't.
        self.create_general_comment(review=self.review_draft)
        self.create_general_comment(
            review=self.review_draft,
            issue_opened=True,
            issue_status=BaseComment.OPEN)

        self.assertSpyNotCalled(on_issue_status_updated)

    def test_issue_updated_signal_with_draft_update(self) -> None:
        """Testing updating the issue status on a draft comment does not
        emit the comment_issue_status_updated signal
        """
        def on_issue_status_updated(**kwargs) -> None:
            pass

        self.addCleanup(comment_issue_status_updated.disconnect,
                        on_issue_status_updated)

        comment = self.create_general_comment(
            review=self.review_draft,
            issue_opened=True,
            issue_status=BaseComment.OPEN)

        comment_issue_status_updated.connect(on_issue_status_updated)
        self.spy_on(on_issue_status_updated)

        comment.issue_status = BaseComment.DROPPED
        comment.save(update_fields=['issue_status'])

        self.assertSpyNotCalled(on_issue_status_updated)

    def test_issue_updated_signal_with_issue_dropped(self) -> None:
        """Testing dropping the issue on a published comment emits the
        comment_issue_status_updated signal
        """
        def on_issue_status_updated(**kwargs) -> None:
            pass

        self.addCleanup(comment_issue_status_updated.disconnect,
                        on_issue_status_updated)

        review_draft = self.review_draft
        comment = self.create_general_comment(
            review=review_draft,
            issue_opened=True,
            issue_status=BaseComment.OPEN)
        review_draft.publish()

        comment_issue_status_updated.connect(on_issue_status_updated)
        self.spy_on(on_issue_status_updated)

        comment.issue_status = BaseComment.DROPPED
        comment.save(update_fields=['issue_status'])

        self.assertSpyCalledWith(on_issue_status_updated,
                                 comment=comment,
                                 prev_status=BaseComment.OPEN,
                                 cur_status=BaseComment.DROPPED)

    def test_issue_updated_signal_with_issue_reopened(self) -> None:
        """Testing re-opening the issue on a published comment emits the
        comment_issue_status_updated signal
        """
        def on_issue_status_updated(**kwargs) -> None:
            pass

        self.addCleanup(comment_issue_status_updated.disconnect,
                        on_issue_status_updated)

        review_draft = self.review_draft
        comment = self.create_general_comment(
            review=review_draft,
            issue_opened=True,
            issue_status=BaseComment.OPEN)
        review_draft.publish()
        comment.issue_status = BaseComment.DROPPED
        comment.save(update_fields=['issue_status'])

        comment_issue_status_updated.connect(on_issue_status_updated)
        self.spy_on(on_issue_status_updated)

        comment.issue_status = BaseComment.OPEN
        comment.save(update_fields=['issue_status'])

        self.assertSpyCalledWith(on_issue_status_updated,
                                 comment=comment,
                                 prev_status=BaseComment.DROPPED,
                                 cur_status=BaseComment.OPEN)

    def test_issue_updated_signal_with_issue_resolved(self) -> None:
        """Testing resolving the issue on a published comment emits the
        comment_issue_status_updated signal
        """
        def on_issue_status_updated(**kwargs) -> None:
            pass

        self.addCleanup(comment_issue_status_updated.disconnect,
                        on_issue_status_updated)

        review_draft = self.review_draft
        comment = self.create_general_comment(
            review=review_draft,
            issue_opened=True,
            issue_status=BaseComment.OPEN)
        review_draft.publish()

        comment_issue_status_updated.connect(on_issue_status_updated)
        self.spy_on(on_issue_status_updated)

        comment.issue_status = BaseComment.RESOLVED
        comment.save(update_fields=['issue_status'])

        self.assertSpyCalledWith(on_issue_status_updated,
                                 comment=comment,
                                 prev_status=BaseComment.OPEN,
                                 cur_status=BaseComment.RESOLVED)



"""Views for SAML SSO.

Version Added:
    5.0
"""

from __future__ import annotations

import logging
from enum import Enum
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.models import User
from django.contrib.auth.views import LoginView
from django.core.cache import cache
from django.http import (Http404,
                         HttpResponse,
                         HttpResponseBadRequest,
                         HttpResponseRedirect,
                         HttpResponseServerError)
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from django.views.generic.base import View
from djblets.cache.backend import make_cache_key
from djblets.db.query import get_object_or_none
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.decorators import cached_property
try:
    from onelogin.saml2.auth import OneLogin_Saml2_Auth
    from onelogin.saml2.errors import OneLogin_Saml2_Error
    from onelogin.saml2.settings import OneLogin_Saml2_Settings
    from onelogin.saml2.utils import OneLogin_Saml2_Utils
except ImportError:
    OneLogin_Saml2_Auth = None
    OneLogin_Saml2_Error = None
    OneLogin_Saml2_Settings = None
    OneLogin_Saml2_Utils = None

from reviewboard.accounts.errors import LoginNotAllowedError
from reviewboard.accounts.models import LinkedAccount
from reviewboard.accounts.sso.backends.saml.forms import SAMLLinkUserForm
from reviewboard.accounts.sso.backends.saml.settings import get_saml2_settings
from reviewboard.accounts.sso.users import (find_suggested_username,
                                            find_user_for_sso_user_id)
from reviewboard.accounts.sso.views import BaseSSOView
from reviewboard.admin.server import get_server_url
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpRequest


logger = logging.getLogger(__name__)


class SAMLViewMixin(View):
    """Mixin to provide common functionality for SAML views.

    Version Added:
        5.0
    """

    def __init__(self, *args, **kwargs):
        """Initialize the view.

        Args:
            *args (tuple):
                Positional arguments to pass through to the base class.

            **kwargs (dict):
                Keyword arguments to pass through to the base class.
        """
        super().__init__(*args, **kwargs)
        self._saml_auth = None
        self._saml_request = None

    def get_saml_request(self, request):
        """Return the SAML request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Information about the SAML request.
        """
        if self._saml_request is None:
            server_url = urlparse(get_server_url())

            if server_url.scheme == 'https':
                https = 'on'
            else:
                https = 'off'

            request_url = request.META['HTTP_HOST']
            http_host = server_url.hostname

            # In some cases, there may be multiple correct server URLs (for
            # example, some users may access the site through a reverse proxy
            # which rewrites all URLs). In this case, we want to allow the SAML
            # auth on those URLs. This requires that ALLOWED_HOSTS is set
            # appropriately to specifically include any URLs that the admin
            # wants.
            if (request_url != server_url.hostname and
                request_url in settings.ALLOWED_HOSTS):
                http_host = request_url

            self._saml_request = {
                'https': https,
                'http_host': http_host,
                'get_data': request.GET.copy(),
                'post_data': request.POST.copy(),
                'query_string': request.META['QUERY_STRING'],
                'request_uri': request.path,
                'script_name': request.META['PATH_INFO'],
                'server_port': server_url.port,
            }

        return self._saml_request

    def get_saml_auth(self, request):
        """Return the SAML auth information.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            onelogin.saml2.auth.OneLogin_Saml2_Auth:
            The SAML Auth object.
        """
        if self._saml_auth is None:
            assert OneLogin_Saml2_Auth is not None
            self._saml_auth = OneLogin_Saml2_Auth(
                self.get_saml_request(request),
                get_saml2_settings())

        return self._saml_auth

    def is_replay_attack(self, message_id):
        """Check for potential replay attacks.

        SAML authentication is potentially vulnerable to a replay attack from a
        man in the middle. This is mitigated by keeping track of recent message
        IDs and rejecting the authentication attempt if we've seen them before.

        Args:
            message_id (str):
                The ID to check.

        Returns:
            bool:
            ``True`` if we've seen the response or assertion IDs before.
            ``False`` if this appears to be a valid request.
        """
        if message_id is None:
            return False

        cache_key = make_cache_key('saml_replay_id_%s' % message_id,
                                   use_encryption=True)
        is_replay = cache.get(cache_key) is not None

        if not is_replay:
            cache.set(cache_key, True)

        return is_replay

    def dispatch(self, *args, **kwargs):
        """Handle a dispatch for the view.

        Args:
            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (dict):
                Keyword arguments to pass through to the parent class.

        Returns:
            django.http.HttpResponse:
            The response to send back to the client.

        Raises:
            django.http.Http404:
                The SAML backend is not enabled, so treat all SAML views as
                404.
        """
        if not self.sso_backend.is_enabled():
            raise Http404

        return super().dispatch(*args, **kwargs)


class SAMLACSView(SAMLViewMixin, BaseSSOView):
    """ACS view for SAML SSO.

    Version Added:
        5.0
    """

    @property
    def success_url(self):
        """The URL to redirect to after a successful login.

        Type:
            str
        """
        url = self.request.POST.get('RelayState')

        assert OneLogin_Saml2_Utils is not None
        self_url = OneLogin_Saml2_Utils.get_self_url(
            self.get_saml_request(self.request))

        if url is not None and self_url != url:
            saml_auth = self.get_saml_auth(self.request)
            return saml_auth.redirect_to(url)
        else:
            return settings.LOGIN_REDIRECT_URL

    @cached_property
    def link_user_url(self):
        """The URL to the link-user flow.

        Type:
            str
        """
        assert self.sso_backend is not None
        return local_site_reverse(
            'sso:%s:link-user' % self.sso_backend.backend_id,
            request=self.request,
            kwargs={'backend_id': self.sso_backend.backend_id})

    def post(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle a POST request.

        Args:
            request (django.http.HttpRequest):
                The request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponse:
            The response to send back to the client.
        """
        auth = self.get_saml_auth(request)
        session = request.session

        try:
            auth.process_response(request_id=session.get('AuthNRequestID'))
        except OneLogin_Saml2_Error as e:
            logger.exception('SAML: Unable to process SSO request: %s', e,
                             exc_info=True, extra={'request': request})
            return HttpResponseBadRequest('Bad SSO response: %s' % str(e),
                                          content_type='text/plain')

        if (self.is_replay_attack(auth.get_last_message_id()) or
            self.is_replay_attack(auth.get_last_assertion_id())):
            logger.error('SAML: Detected replay attack',
                         extra={'request': request})
            return HttpResponseBadRequest(
                'SAML message IDs have already been used')

        error = auth.get_last_error_reason()

        if error:
            logger.error('SAML: Unable to process SSO request: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest('Bad SSO response: %s' % error,
                                          content_type='text/plain')

        # Store some state on the session to identify where we are in the SAML
        # workflow.
        session.pop('AuthNRequestID', None)

        linked_account = get_object_or_none(LinkedAccount,
                                            service_id='sso:saml',
                                            service_user_id=auth.get_nameid())

        if linked_account:
            user = linked_account.user

            try:
                self.sso_backend.login_user(request, user)
            except LoginNotAllowedError:
                return render(request=request,
                              template_name='permission_denied.html',
                              status=403)

            return HttpResponseRedirect(self.success_url)
        else:
            username = auth.get_nameid()

            siteconfig = SiteConfiguration.objects.get_current()
            attr_email = siteconfig.get('saml_attr_email')
            attr_firstname = siteconfig.get('saml_attr_firstname')
            attr_lastname = siteconfig.get('saml_attr_lastname')

            try:
                email = self._get_user_attr_value(auth, attr_email)
                first_name = self._get_user_attr_value(auth, attr_firstname)
                last_name = self._get_user_attr_value(auth, attr_lastname)
            except KeyError as e:
                logger.error('SAML: Assertion is missing %s attribute', e,
                             extra={'request': request})
                return HttpResponseBadRequest('Bad SSO response: assertion is '
                                              'missing %s attribute'
                                              % e,
                                              content_type='text/plain')

            request.session['sso'] = {
                'user_data': {
                    'id': username,
                    'first_name': first_name,
                    'last_name': last_name,
                    'email': email,
                },
                'raw_user_attrs': auth.get_attributes(),
                'session_index': auth.get_session_index(),
            }

            return HttpResponseRedirect(self.link_user_url)

    def _get_user_attr_value(self, auth, key):
        """Return the value of a user attribute.

        Args:
            auth (onelogin.saml2.auth.OneLogin_Saml2_Auth):
                The SAML authentication object.

            key (str):
                The key to look up.

        Returns:
            str:
            The attribute, if it exists.

        Raises:
            KeyError:
                The given key was not present in the SAML assertion.
        """
        value = auth.get_attribute(key)

        if value and isinstance(value, list):
            return value[0]

        siteconfig = SiteConfiguration.objects.get_current()
        attr_firstname = siteconfig.get('saml_attr_firstname')
        attr_lastname = siteconfig.get('saml_attr_lastname')

        # Some identity providers only allow setting the full name, not
        # separate first and last. In this case, we need to fake it by
        # splitting.
        if key in (attr_firstname, attr_lastname):
            try:
                attr_fullname = siteconfig.get('saml_attr_fullname')
                fullname = self._get_user_attr_value(auth, attr_fullname)
            except KeyError:
                # Reraise with the original key name so that this fallback
                # isn't exposed in the exception message.
                raise KeyError(key)

            # We don't have a good way to split the user's full name to a first
            # and last name, so we split on the first space and then treat the
            # two parts as the first and last names.
            name_parts = fullname.split(' ', 1)

            if key == attr_firstname:
                return name_parts[0]
            else:
                return len(name_parts) > 1 and name_parts[1] or ''

        raise KeyError(key)


@method_decorator(csrf_protect, name='dispatch')
class SAMLLinkUserView(SAMLViewMixin, BaseSSOView, LoginView):
    """Link user view for SAML SSO.

    This can have several behaviors depending on what combination of state we
    get from the Identity Provider and what we have stored in the database.

    The first major case is where we are given data that matches an existing
    user in the database. Ideally this is via the "username" field, but may
    also be a matching e-mail address, or parsing a username out of the e-mail
    address.

    In this case, there are two paths. The simple path is where the
    administrator trusts both the authority and integrity of their Identity
    Provider and has turned off the "Require login to link" setting. For this,
    we'll just create the LinkedAccount, authenticate the user, and redirect to
    the success URL.

    If the require login setting is turned on, the user will have a choice.
    They can enter the password for the detected user to complete the link. If
    they have an account but the detected one is not correct, they can log in
    with their username and password to link the other account. Finally, they
    can provision a new user if they do not yet have one.

    The second major case is where we cannot find an existing user. In this
    case, we'll offer the user a choice: if they have an existing login that
    wasn't found, they can log in with their (non-SSO) username and password.
    If they don't have an account, they will be able to provision one.

    Version Added:
        5.0
    """

    # TODO: This has a lot of logic which will likely be applicable to other
    # SSO backend implementations. When we add a new backend, this should be
    # refactored to pull out most of the logic into a common base class, and
    # just implement SAML-specific data here.

    form_class = SAMLLinkUserForm

    class Mode(Enum):
        CONNECT_EXISTING_ACCOUNT = 'connect'
        CONNECT_WITH_LOGIN = 'connect-login'
        PROVISION = 'provision'

    def dispatch(self, *args, **kwargs):
        """Dispatch the view.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        self._sso_user_data = \
            self.request.session.get('sso', {}).get('user_data')
        self._sso_data_username = self._sso_user_data.get('id')
        self._sso_data_email = self._sso_user_data.get('email')
        computed_username = find_suggested_username(self._sso_data_email)
        self._provision_username = self._sso_data_username or computed_username
        self._sso_user = find_user_for_sso_user_id(
            self._sso_data_username,
            self._sso_data_email,
            computed_username)

        requested_mode = self.request.GET.get('mode')
        self._mode = None

        if requested_mode:
            try:
                self._mode = self.Mode(requested_mode)
            except ValueError:
                pass

        if self._mode is None:
            if self._sso_user:
                self._mode = self.Mode.CONNECT_EXISTING_ACCOUNT
            else:
                self._mode = self.Mode.PROVISION

        return super(SAMLLinkUserView, self).dispatch(*args, **kwargs)

    def get_template_names(self):
        """Return the template to use when rendering.

        Returns:
            list:
            A single-item list with the template name to use when rendering.

        Raises:
            ValueError:
                The current mode is not valid.
        """
        if self._mode == self.Mode.CONNECT_EXISTING_ACCOUNT:
            return ['accounts/sso/link-user-connect-existing.html']
        elif self._mode == self.Mode.CONNECT_WITH_LOGIN:
            return ['accounts/sso/link-user-login.html']
        elif self._mode == self.Mode.PROVISION:
            siteconfig = SiteConfiguration.objects.get_current()

            if siteconfig.get('saml_automatically_provision_users', True):
                return ['accounts/sso/link-user-provision.html']
            else:
                return ['accounts/sso/link-user-provision-disabled.html']
        else:
            raise ValueError('Unknown link-user mode "%s"' % self._mode)

    def get_initial(self):
        """Return the initial data for the form.

        Returns:
            dict:
            Initial data for the form.
        """
        initial = super(SAMLLinkUserView, self).get_initial()

        if self._sso_user is not None:
            initial['username'] = self._sso_user.username
        else:
            initial['username'] = self._provision_username

        initial['provision'] = (self._mode == self.Mode.PROVISION)

        return initial

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template.

        Args:
            **kwargs (dict):
                Keyword arguments for the view.

        Returns:
            dict:
            Additional data to inject into the render context.
        """
        context = super(SAMLLinkUserView, self).get_context_data(**kwargs)
        context['user'] = self._sso_user
        context['mode'] = self._mode
        context['username'] = self._provision_username

        return context

    def get(self, request, *args, **kwargs):
        """Handle a GET request for the form.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments to pass through to the base class.

            **kwargs (dict):
                Keyword arguments to pass through to the base class.

        Returns:
            django.http.HttpResponse:
            The response to send back to the client.
        """
        if not self._sso_user_data:
            return HttpResponseRedirect(
                local_site_reverse('login', request=request))

        siteconfig = SiteConfiguration.objects.get_current()

        if self._sso_user and not siteconfig.get('saml_require_login_to_link'):
            return self.link_user(self._sso_user)

        return super(SAMLLinkUserView, self).get(request, *args, **kwargs)

    def form_valid(
        self,
        form: SAMLLinkUserForm,
    ) -> HttpResponse:
        """Handle a successfully authenticated form.

        Args:
            form (reviewboard.accounts.sso.backends.saml.forms.
                  SAMLLinkUserForm):
                The link-user form.

        Returns:
            django.http.HttpResponseRedirect:
            A redirect to the next page.
        """
        if form.cleaned_data['provision']:
            # We can't provision if there's an existing matching user.
            # TODO: show an error?
            assert not self._sso_user
            assert self._provision_username

            first_name = self._sso_user_data.get('first_name')
            last_name = self._sso_user_data.get('last_name')

            logger.info('SAML: Provisioning user "%s" (%s <%s %s>)',
                        self._provision_username, self._sso_data_email,
                        first_name, last_name, extra={'request': self.request})

            user = User.objects.create(
                username=self._provision_username,
                email=self._sso_data_email,
                first_name=first_name,
                last_name=last_name)
        else:
            user = form.get_user()

        return self.link_user(user)

    def link_user(
        self,
        user: User,
    ) -> HttpResponse:
        """Link the given user.

        Args:
            user (django.contrib.auth.models.User):
                The user to link.

        Returns:
            django.http.HttpResponse:
            A redirect to the success URL, or an error page.
        """
        sso_id = self._sso_user_data.get('id')

        logger.info('SAML: Linking SSO user "%s" to Review Board user "%s"',
                    sso_id, user.username, extra={'request': self.request})

        user.linked_accounts.create(
            service_id='sso:saml',
            service_user_id=sso_id)

        self.sso_backend.login_user(self.request, user)

        return HttpResponseRedirect(self.get_success_url())


class SAMLLoginView(SAMLViewMixin, BaseSSOView):
    """Login view for SAML SSO.

    Version Added:
        5.0
    """

    def get(self, request, *args, **kwargs):
        """Handle a GET request for the login URL.

        Args:
            request (django.http.HttpRequest):
                The request from the client.

            *args (tuple, unused):
                Additional positional arguments.

            **kwargs (dict, unused):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponseRedirect:
            A redirect to start the login flow.
        """
        auth = self.get_saml_auth(request)

        redirect = self.request.GET.get(REDIRECT_FIELD_NAME,
                                        settings.LOGIN_REDIRECT_URL)

        return HttpResponseRedirect(auth.login(redirect))


class SAMLMetadataView(SAMLViewMixin, BaseSSOView):
    """Metadata view for SAML SSO.

    Version Added:
        5.0
    """

    def get(self, request, *args, **kwargs):
        """Handle a GET request.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments from the URL definition.

            **kwargs (dict):
                Keyword arguments from the URL definition.
        """
        assert OneLogin_Saml2_Settings is not None
        saml_settings = OneLogin_Saml2_Settings(
            get_saml2_settings(),
            sp_validation_only=True)

        metadata = saml_settings.get_sp_metadata()
        errors = saml_settings.validate_metadata(metadata)

        if errors:
            logger.error('SAML: Got errors from metadata validation: %s',
                         ', '.join(errors),
                         extra={'request': request})
            return HttpResponseServerError(', '.join(errors),
                                           content_type='text/plain')

        return HttpResponse(metadata, content_type='text/xml')


class SAMLSLSView(SAMLViewMixin, BaseSSOView):
    """SLS view for SAML SSO.

    Version Added:
        5.0
    """

    def get(self, request, *args, **kwargs):
        """Handle a GET request.

        Args:
            request (django.http.HttpRequest):
                The request from the client.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.http.HttpResponse:
            The response to send back to the client.
        """
        auth = self.get_saml_auth(request)
        request_id = None

        if 'LogoutRequestId' in request.session:
            request_id = request.session['LogoutRequestId']

        redirect_url = auth.process_slo(
            request_id=request_id,
            delete_session_cb=lambda: request.session.flush())

        if (self.is_replay_attack(auth.get_last_message_id()) or
            self.is_replay_attack(auth.get_last_request_id())):
            logger.error('SAML: Detected replay attack',
                         extra={'request': request})
            return HttpResponseBadRequest(
                'SAML message IDs have already been used')

        error = auth.get_last_error_reason()

        if error:
            logger.error('SAML: Unable to process SLO request: %s', error,
                         extra={'request': request})
            return HttpResponseBadRequest('Bad SLO response: %s' % error,
                                          content_type='text/plain')

        if redirect_url:
            return HttpResponseRedirect(redirect_url)
        else:
            return HttpResponseRedirect(settings.LOGIN_URL)



"""Tests for reviewboard.diffviewer.filediff_creator."""

from __future__ import annotations

import unittest
from typing import TYPE_CHECKING, Union

import kgb
from django.utils.timezone import now

from reviewboard.diffviewer.filediff_creator import create_filediffs
from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.scmtools.core import Revision
from reviewboard.scmtools.git import GitTool
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from reviewboard.scmtools.core import FileLookupContext


class FileDiffCreatorTests(kgb.SpyAgency, TestCase):
    """Tests for reviewboard.diffviewer.filediff_creator."""

    fixtures = ['test_scmtools']

    def test_create_filediffs_file_count(self):
        """Testing create_filediffs() with a DiffSet"""
        repository = self.create_repository()
        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        create_filediffs(
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)

    def test_create_filediffs_commit_file_count(self):
        """Testing create_filediffs() with a DiffSet and a DiffCommit"""
        repository = self.create_repository()
        diffset = DiffSet.objects.create_empty(repository=repository)
        commits = [
            DiffCommit.objects.create(
                diffset=diffset,
                filename='diff',
                author_name='Author Name',
                author_email='author@example.com',
                commit_message='Message',
                author_date=now(),
                commit_id='a' * 40,
                parent_id='0' * 40),
            DiffCommit.objects.create(
                diffset=diffset,
                filename='diff',
                author_name='Author Name',
                author_email='author@example.com',
                commit_message='Message',
                author_date=now(),
                commit_id='b' * 40,
                parent_id='a' * 40),
        ]

        self.assertEqual(diffset.files.count(), 0)
        self.assertEqual(commits[0].files.count(), 0)

        create_filediffs(
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            diffcommit=commits[0],
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)
        commits[0] = DiffCommit.objects.get(pk=commits[0].pk)

        self.assertEqual(diffset.files.count(), 1)
        self.assertEqual(commits[0].files.count(), 1)

        create_filediffs(
            diff_file_contents=self.DEFAULT_GIT_FILEDIFF_DATA_DIFF,
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            diffcommit=commits[1],
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)
        commits[1] = DiffCommit.objects.get(pk=commits[1].pk)

        self.assertEqual(diffset.files.count(), 2)
        self.assertEqual(commits[1].files.count(), 1)

    def test_create_filediffs_with_symlinks(self):
        """Testing create_filediffs() with symlinks"""
        repository = self.create_repository(tool_name='TestToolDiffX')
        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        create_filediffs(
            diff_file_contents=(
                b'#diffx: encoding=utf-8, version=1.0\n'
                b'#.change:\n'
                b'#..file:\n'
                b'#...meta: format=json, length=140\n'
                b'{\n'
                b'    "op": "modify",\n'
                b'    "path": "name",\n'
                b'    "revision": {\n'
                b'        "old": "abc123",\n'
                b'        "new": "def456"\n'
                b'    },\n'
                b'    "type": "symlink"\n'
                b'}\n'
            ),
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)
        filediff = diffset.files.get()

        self.assertTrue(filediff.is_symlink)
        self.assertIsNone(filediff.old_symlink_target)
        self.assertIsNone(filediff.new_symlink_target)

    def test_create_filediffs_with_symlinks_and_targets(self):
        """Testing create_filediffs() with symlinks and symlink targets"""
        repository = self.create_repository(tool_name='TestToolDiffX')
        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        create_filediffs(
            diff_file_contents=(
                b'#diffx: encoding=utf-8, version=1.0\n'
                b'#.change:\n'
                b'#..file:\n'
                b'#...meta: format=json, length=230\n'
                b'{\n'
                b'    "op": "modify",\n'
                b'    "path": "name",\n'
                b'    "revision": {\n'
                b'        "old": "abc123",\n'
                b'        "new": "def456"\n'
                b'    },\n'
                b'    "symlink target": {\n'
                b'        "old": "old/target/",\n'
                b'        "new": "new/target/"\n'
                b'    },\n'
                b'    "type": "symlink"\n'
                b'}\n'
            ),
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)
        filediff = diffset.files.get()

        self.assertTrue(filediff.is_symlink)
        self.assertEqual(filediff.old_symlink_target, 'old/target/')
        self.assertEqual(filediff.new_symlink_target, 'new/target/')

    def test_create_filediffs_with_unix_mode(self):
        """Testing create_filediffs() with UNIX file modes"""
        repository = self.create_repository(tool_name='TestToolDiffX')
        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        create_filediffs(
            diff_file_contents=(
                b'#diffx: encoding=utf-8, version=1.0\n'
                b'#.change:\n'
                b'#..file:\n'
                b'#...meta: format=json, length=199\n'
                b'{\n'
                b'    "op": "modify",\n'
                b'    "path": "name",\n'
                b'    "revision": {\n'
                b'        "old": "abc123",\n'
                b'        "new": "def456"\n'
                b'    },\n'
                b'    "unix file mode": {\n'
                b'        "old": "0100644",\n'
                b'        "new": "0100755"\n'
                b'    }\n'
                b'}\n'
            ),
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)
        filediff = diffset.files.get()

        self.assertEqual(filediff.old_unix_mode, '0100644')
        self.assertEqual(filediff.new_unix_mode, '0100755')

    def test_create_filediffs_with_parent_and_revision_instance(self) -> None:
        """Testing create_filediffs() with parent diff and revisions as
        Revision
        """
        repository = self.create_repository(tool_name='Git')
        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        def _parse_diff_revision(
            self: object,
            filename: bytes,
            revision: bytes,
            *args,
            **kwargs,
        ) -> tuple[bytes, Union[Revision, bytes]]:
            return (filename, Revision(revision.decode('utf-8')))

        # Make sure we run this test with a Revision as a result.
        self.spy_on(GitTool.parse_diff_revision,
                    owner=GitTool,
                    call_fake=_parse_diff_revision)

        # We'll use the same diff for both tests. It doesn't really matter.
        # We're just looking for the end result of the parsed filenames and
        # revisions here.
        create_filediffs(
            diff_file_contents=(
                b'diff --git a/readme b/readme\n'
                b'index 1234567..7654321\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1 @@\n'
                b'Test 1\n'
                b'Test 2\n'
            ),
            parent_diff_file_contents=(
                b'diff --git /dev/null b/readme\n'
                b'index 1234567..7654321\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1 @@\n'
                b'Test 1\n'
                b'Test 2\n'
            ),
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            check_existence=False)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)
        filediff = diffset.files.get()

        self.assertEqual(filediff.source_revision, '1234567')
        self.assertEqual(filediff.dest_detail, '7654321')
        self.assertEqual(filediff.extra_data.get('parent_source_filename'),
                         '/readme')
        self.assertEqual(filediff.extra_data.get('parent_source_revision'),
                         '1234567')

    def test_create_filediffs_without_per_file_revisions(self) -> None:
        """Testing create_filediffs() without per-file revisions and
        validated parent ID
        """
        repository = self.create_repository(tool_name='Mercurial')

        try:
            repository.get_scmtool()
        except ImportError:
            raise unittest.SkipTest('Hg is not installed')

        diffset = self.create_diffset(repository=repository)

        self.assertEqual(diffset.files.count(), 0)

        def get_file_exists(
            *args,
            context: FileLookupContext,
            **kwargs,
        ) -> bool:
            context.file_extra_data['__validated_parent_id'] = 'zzzzzzzzzzzz'

            return True

        create_filediffs(
            diff_file_contents=(
                b'# HG changeset patch\n'
                b'# Node ID aaaaaaaaaaaa\n'
                b'# Parent  bbbbbbbbbbbb\n'
                b'diff --git a/readme b/readme\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1 @@\n'
                b'Test 1\n'
                b'Test 2\n'
            ),
            parent_diff_file_contents=None,
            repository=repository,
            basedir='/',
            base_commit_id='0' * 40,
            diffset=diffset,
            get_file_exists=get_file_exists)

        diffset = DiffSet.objects.get(pk=diffset.pk)

        self.assertEqual(diffset.files.count(), 1)
        filediff = diffset.files.get()

        self.assertEqual(filediff.source_revision, 'zzzzzzzzzzzz')
        self.assertEqual(filediff.dest_detail, 'aaaaaaaaaaaa')
        self.assertEqual(filediff.extra_data, {
            'is_symlink': False,
            'raw_delete_count': 0,
            'raw_insert_count': 0,
        })



"""Manually add an index for ReviewRequest.summary.

MySQL has a quirk (many say a bug) where the max index length is based on the
actual number of allowed bytes, rather than characters, in a VARCHAR. On
InnoDB, this max length is 767 bytes. Each Unicode character in MySQL takes 3
bytes, though. If you're running InnoDB with utf-8 and create an index on a
column with a max length over 255 characters (or 765 bytes, you'll either get
an index safely capped to 255 (if running MySQL < 5.6) or an error (if running
MySQL 5.6+).

Note that this is not a problem for Postgres or SQLite. It's also, in this
case, not a problem for MySQL with MyISAM, since the max length there is 1000
bytes.

Our options for the summary field were to either truncate from 300 characters
to 255 (destroying bits of history), to remove the index entirely, or to cap
the index ourselves. The first two options are terrible, and the third is not
directly supported by Django.

Turns out what we can do is create per-engine SQL initial data files that will
create the appropriate indexes. This will only work for new installations,
though, but covers that case.

This evolution then, depending on the database type, invokes custom SQL for
creating the index. On MySQL, that custom SQL will contain a key length cap,
and for everything else, there will be no cap (since they don't support such a
concept).

Version Added:
    2.0.11
"""

from django.conf import settings
from django_evolution.mutations import ChangeField, SQLMutation


if settings.DATABASES['default']['ENGINE'].endswith('mysql'):
    field_index_suffix = '(255)'
else:
    field_index_suffix = ''

index_sql = (
    'CREATE INDEX reviews_reviewrequest_summary ON reviews_reviewrequest'
    ' (summary%s);'
    % field_index_suffix
)


MUTATIONS = [
    ChangeField('ReviewRequest', 'summary', initial=None, db_index=False),
    SQLMutation('reviewrequest_summary', [index_sql]),
]



"""Unit tests for AdminSidebarManageActionGroupRenderer.

Version Added:
    8.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import kgb
from django.contrib.auth.models import User
from django.core.cache import cache
from django.db.models import Count
from django.template import Context
from django_assert_queries import assert_queries
from djblets.cache.backend import make_cache_key

from reviewboard.actions import actions_registry
from reviewboard.actions.base import ActionPlacement, BaseGroupAction
from reviewboard.actions.tests.base import TestActionsRegistry
from reviewboard.admin.actions import (AdminManageNavGroupAction,
                                       AdminSidebarManageActionGroupRenderer,
                                       BaseAdminSidebarManageItemAction)
from reviewboard.testing.testcase import TestCase

if TYPE_CHECKING:
    from django.db.models import Model


class BaseAdminSidebarManageItemActionTests(kgb.SpyAgency, TestCase):
    """Unit tests for AdminSidebarManageActionGroupRenderer.

    Version Added:
        8.0
    """

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        # Rendering any admin page stamps the managed-models set onto the
        # registered manage group action, and that survives across tests.
        # Clear it so these tests fully control the state.
        action = actions_registry.get_action(
            AdminManageNavGroupAction.action_id)

        if action is not None:
            action.__dict__.pop('_managed_models', None)

    def test_get_extra_context(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer.get_extra_context"""
        class MyManageGroupAction(BaseGroupAction):
            action_id = 'my-group'
            label = 'My Group'

            placements = [
                ActionPlacement(attachment='test'),
            ]

        class MyManageItemAction(BaseAdminSidebarManageItemAction):
            action_id = 'my-action'
            label = 'My Action'
            add_item_url_name = 'admin:auth_user_add'
            item_queryset = User.objects.all()

            placements = [
                ActionPlacement(attachment='test',
                                parent_id=MyManageGroupAction.action_id),
            ]

        group_action = MyManageGroupAction()
        item_action = MyManageItemAction()

        test_registry = TestActionsRegistry()
        test_registry.register(group_action)
        test_registry.register(item_action)

        self.create_user(username='test')

        request = self.create_http_request()
        placement = group_action.get_placement('test')
        renderer = AdminSidebarManageActionGroupRenderer(
            action=group_action,
            placement=placement)

        equeries = [
            {
                'annotations': {
                    '__count': Count('*'),
                },
                'model': User,
            },
        ]

        with assert_queries(equeries):
            extra_context = renderer.get_extra_context(
                request=request,
                context=Context({
                    'request': request,
                }))

        self.assertEqual(
            extra_context,
            {
                'action': group_action,
                'action_renderer': renderer,
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'attachment_point_id': 'test',
                'children': [
                    item_action,
                ],
                'dom_element_id': 'action-test-my-group',
                'has_parent': False,
                'id': 'my-group',
                'is_toplevel': True,
                'item_counts': {
                    'my-action': 1,
                },
                'label': 'My Group',
                'placement': placement,
                'url': '#',
                'verbose_label': None,
                'visible': True,
            })

        self.assertEqual(
            cache.get(make_cache_key('admin-sidebar-manage-state')),
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # A second call should reuse cache.
        with self.assertNumQueries(0):
            extra_context = renderer.get_extra_context(
                request=request,
                context=Context({
                    'request': request,
                }))

        self.assertEqual(
            extra_context,
            {
                'action': group_action,
                'action_renderer': renderer,
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'attachment_point_id': 'test',
                'children': [
                    item_action,
                ],
                'dom_element_id': 'action-test-my-group',
                'has_parent': False,
                'id': 'my-group',
                'is_toplevel': True,
                'item_counts': {
                    'my-action': 1,
                },
                'label': 'My Group',
                'placement': placement,
                'url': '#',
                'verbose_label': None,
                'visible': True,
            })

    def test_cache_invalidate_after_save_new_managed(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        invalidates after saving new managed model
        """
        self._patch_managed_models({User})

        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Save a new object.
        self.create_user(username='test1')

        self.assertIsNone(cache.get(key))

    def test_cache_invalidate_after_save_new_unmanaged(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        does not invalidate after saving new unmanaged model
        """
        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Save a new object.
        self.create_user(username='test1')

        self.assertEqual(
            cache.get(key),
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

    def test_cache_invalidate_after_delete_managed(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        invalidates after delete of managed model
        """
        self._patch_managed_models({User})

        user = self.create_user(username='test1')

        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Delete the object.
        user.delete()

        self.assertIsNone(cache.get(key))

    def test_cache_invalidate_after_delete_unmanaged(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        does not invalidate after delete of unmanaged model
        """
        user = self.create_user(username='test1')

        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Delete the object.
        user.delete()

        self.assertEqual(
            cache.get(key),
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

    def test_cache_preserves_after_save_update_managed(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        does not invalidate after saving an existing managed model
        """
        self._patch_managed_models({User})

        user = self.create_user(username='test1')

        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Update the object.
        user.save()

        self.assertEqual(
            cache.get(key),
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

    def test_cache_preserves_after_save_update_unmanaged(self) -> None:
        """Testing AdminSidebarManageActionGroupRenderer item counts cache
        does not invalidate after saving an existing unmanaged model
        """
        user = self.create_user(username='test1')

        key = make_cache_key('admin-sidebar-manage-state')

        cache.add(
            key,
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

        # Update the object.
        user.save()

        self.assertEqual(
            cache.get(key),
            {
                'add_item_urls': {
                    'my-action': '/admin/db/auth/user/add/',
                },
                'item_counts': {
                    'my-action': 1,
                },
            })

    def _patch_managed_models(
        self,
        models: set[type[Model]],
    ) -> None:
        """Patch managed models into the manage group state.

        These will be cleaned up when the test ends.

        Args:
            models (set of type):
                The models to patch in as managed models.
        """
        # Patch the registered action instance, since that's what the cache
        # invalidation signal handlers consult. An instance attribute would
        # shadow anything set on the class.
        action = actions_registry.get_action(
            AdminManageNavGroupAction.action_id)
        assert action is not None

        setattr(action, '_managed_models', models)
        self.addCleanup(
            lambda: action.__dict__.pop('_managed_models', None))
